/*---------------------------------------------------------------------------*/
/* GIVE-AUTH.C -- contains authorization control routines for GIVE.C         */
/*---------------------------------------------------------------------------*/

#include <jpidef.h>
#include <ssdef.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <str$routines.h>
#include <strdef.h>
#include "give.h"
#include "give-jpi.h"
#include "give-auth.h"
#include "give-str.h"
#include "give-rms.h"

/*---------------------------------------------------------------------------*/
/* Check_Authorization -- checks to see if the running user is authorized to */
/*  give the file to the destination user.                                   */

ulong Check_Authorization( char *dest_directory, ulong file_size,
                           ulong diskquota, char *actual_dest_filename )
{
   ulong ss_stat;
   ulong i;

   char my_username[K_USERNAME_LEN+1];
   $DESCRIPTOR( my_username_d, my_username );
   char my_account[K_ACCOUNT_LEN+1];
   $DESCRIPTOR( my_account_d, my_account );
   char hold_string[256];
   $DESCRIPTOR( hold_string_d, hold_string );
   char auth_filename[K_FILENAME_LEN+1];
   ulong rights_size;
   struct rights_entry *my_rights;
   struct dsc$descriptor_s *my_rights_d;
   
   struct auth_entry *auth_head;
   struct auth_entry *curr_node;

/*------*/
/* See if the file would put the user within 100 blocks of disk quota        */

   if( file_size + 100 > diskquota )
      return( GIVE_AUTH_EXQUOTA );
 
/*------*/
/* Construct authorization filename                                          */

   strcpy( auth_filename, dest_directory );
   strcat( auth_filename, K_AUTH_FILENAME );

   if( qual_debug )
      printf( "Authorization: %s\n", auth_filename );

/*------*/
/* Read authorization file                                                   */

   ss_stat = Read_Authorization_File( auth_filename, &auth_head );
   if( ss_failure( ss_stat ) )
      return( ss_stat );

   if( qual_debug )
      printf( "Authorization read.\n" );

/*------*/
/* Get running user's account information.                                   */

   ss_stat = Get_Running_User_Information( &my_username_d, &my_account_d,
                                           &my_rights_d );
   if( ss_failure( ss_stat ) )
      return( ss_stat );

   if( qual_debug ) {
      printf( "My username: %s.\n", my_username_d.dsc$a_pointer );
      printf( "My account:  %s.\n", my_account_d.dsc$a_pointer );
      printf( "My rights:   " );
      for( i=0; my_rights_d[i].dsc$a_pointer != NULL; i++ ) {
         if( i>0 ) printf( "," );
         printf( "%s",my_rights_d[i].dsc$a_pointer );
      }
      printf( "\n" );
   }

/*------*/
/* Add the directory offset to the destination directory.                    */

   dest_directory[strlen(dest_directory)-1]='\0';
   strcat( dest_directory, K_DESTDIR_POSTFIX );

/*------*/
/* Now scan through the authorization list.  If an entry matches, and has    */
/* all of the appropriate characteristics, return a success.  Otherwise,     */
/* return a failure.                                                         */

   ss_stat = Look_For_Authorization( auth_head, dest_directory, file_size,
                                     actual_dest_filename, my_username_d, 
                                     my_account_d, my_rights_d );
   if( ss_failure( ss_stat ) )
      return( ss_stat );

   return( SS$_NORMAL );
}

/*---------------------------------------------------------------------------*/
/* Get_Running_User_Information -- gets information about the user running   */
/*   the program.                                                            */

ulong Get_Running_User_Information( struct dsc$descriptor_s *username_d, 
                                    struct dsc$descriptor_s *account_d,
                                    struct dsc$descriptor_s **rights_d )
{

   ulong ss_stat;
   ulong i;

   char hold_string[256];
   $DESCRIPTOR( hold_string_d, hold_string );
   ulong rights_size;
   struct rights_entry *rights;
   
/*------*/
/* Get the username.  Strip off trailing spaces.                             */

   ss_stat = Get_Single_JPI( 0, JPI$_USERNAME, sizeof( hold_string ),
                             hold_string, &(hold_string_d.dsc$w_length) );
   if( ss_failure( ss_stat ) )
      return( ss_stat );
   ss_stat = str$trim( username_d, &hold_string_d,
                       &(username_d->dsc$w_length) );
   if( ss_failure( ss_stat ) )
      return( ss_stat );
   null_terminate( username_d->dsc$a_pointer, username_d->dsc$w_length );

/*------*/
/* Get the account name.                                                     */

   hold_string_d.dsc$w_length = sizeof( hold_string );
   ss_stat = Get_Single_JPI( 0, JPI$_ACCOUNT, sizeof( hold_string ),
                             hold_string, &(hold_string_d.dsc$w_length) );
   if( ss_failure( ss_stat ) )
      return( ss_stat );
   ss_stat = str$trim( account_d, &hold_string_d,
                       &(account_d->dsc$w_length) );
   if( ss_failure( ss_stat ) )
      return( ss_stat );
   null_terminate( account_d->dsc$a_pointer, account_d->dsc$w_length );

/*------*/
/* Get the rights identifiers.  First find out how big we expect the rights  */
/* list to be, double that number for safety (!), and allocate an array of   */
/* structures that size.  Count this and allocate an array to return         */
/* pointers to the names in.  Convert all the identifiers to text, and dup   */
/* the strings into the newly allocated return array.  Put a NULL at the     */
/* end of this array, and free up the intermediate identifier-structure      */
/* array.                                                                    */

   ss_stat = Get_Single_JPI( 0, JPI$_RIGHTS_SIZE, sizeof( rights_size ),
                             &rights_size, NULL );
   if( ss_failure( ss_stat ) )
      return( ss_stat );
   rights_size <<= 1;
   if( (rights = malloc( rights_size ))==NULL )
      return( GIVE_AUTH_MEMERR );
   ss_stat = Get_Single_JPI( 0, JPI$_PROCESS_RIGHTS, rights_size,
                             rights, NULL );
   if( ss_failure( ss_stat ) )
      return( ss_stat );
   
   for( i=0; rights[i].identifier != 0; i++ );

   if( (*rights_d = malloc( (i+1)*sizeof( struct dsc$descriptor_s )))==NULL )
      return( GIVE_AUTH_MEMERR );

   for( i=0; rights[i].identifier != 0; i++ ) {
      hold_string_d.dsc$w_length = sizeof( hold_string );
      ss_stat = sys$idtoasc( rights[i].identifier,
                             &(hold_string_d.dsc$w_length),
                             &hold_string_d, 0, 0, 0 );
      if( ss_failure( ss_stat ) )
         return( ss_stat );
      null_terminate( hold_string, hold_string_d.dsc$w_length );
      if( ((*rights_d)[i].dsc$a_pointer = strdup( hold_string ))==NULL )
         return( GIVE_AUTH_MEMERR );
      (*rights_d)[i].dsc$b_dtype = DSC$K_DTYPE_T;
      (*rights_d)[i].dsc$b_class = DSC$K_CLASS_S;
      (*rights_d)[i].dsc$w_length = hold_string_d.dsc$w_length;
   }
   (*rights_d)[i].dsc$a_pointer = NULL;

   free( rights );

   return( SS$_NORMAL );

}

/*---------------------------------------------------------------------------*/
/* Read_Authorization_File -- read the authorization file and build a linked */
/* list containing its contents.                                             */

ulong Read_Authorization_File( char *filename, struct auth_entry **auth_list )
{
   ulong ss_stat;
   ulong i;
   char auth_line[K_AUTHLINE_LEN+1];
   $DESCRIPTOR( auth_line_d, auth_line );
   char trim_line[K_AUTHLINE_LEN+1];
   $DESCRIPTOR( trim_line_d, trim_line );
   FILE *auth_fp;
   struct auth_entry *curr_node;
   char *token;

/*------*/
/* Allocate and set up the header node                                       */

   if( (curr_node = malloc( sizeof( struct auth_entry )))==NULL )
      return( GIVE_AUTH_MEMERR );
   curr_node->flags = K_INCOMING_HEADER;
   *auth_list = curr_node;

/*------*/
/* Open the authorization file                                               */

   if( (auth_fp = fopen( filename, "r" ))==NULL )
      return( GIVE_AUTH_OPENERR );

/*------*/
/* Loop through file, reading lines                                          */

   while( fgets( auth_line, sizeof( auth_line )-1, auth_fp ) != NULL ) {

   /*------*/
   /* Allocate the next node in the list                                     */

      if( (curr_node->next = malloc( sizeof( struct auth_entry )))==NULL )
         return( GIVE_AUTH_MEMERR );
      curr_node->next->next = NULL;

   /*------*/
   /* Truncate the string at the first comment.  Make sure the string ends   */
   /* with an newline.                                                       */

      for( i=0; auth_line[i] != '\0'; i++ ) {
         if( auth_line[i] == K_AUTH_COMMENT_CHAR ) {
            auth_line[i] = '\0';
            break;
         }
      }
      strcat( auth_line, "\n" );

   /*------*/
   /* Get the first token, the username.  Failure to find the first token    */
   /* is not an error, but rather indicates a blank line and causes a skip   */
   /* to the beginning of the next iteration.                                */

      token = strtok( auth_line, K_AUTH_TOKEN_SEPARATORS );
      if( token == NULL )
         continue;
      strupr( token );
      if( (curr_node->next->username.dsc$a_pointer = strdup( token ))==NULL )
         return( GIVE_AUTH_MEMERR );
      curr_node->next->username.dsc$w_length =
         strlen( curr_node->next->username.dsc$a_pointer );
      curr_node->next->username.dsc$b_dtype = DSC$K_DTYPE_T;
      curr_node->next->username.dsc$b_class = DSC$K_CLASS_S;

   /*------*/
   /* Get the second token, the identifier.                                  */

      token = strtok( NULL, K_AUTH_TOKEN_SEPARATORS );
      if( token == NULL )
         return( GIVE_AUTH_PARSERR );
      strupr( token );
      if( (curr_node->next->identifier.dsc$a_pointer = strdup( token ))==NULL )
         return( GIVE_AUTH_MEMERR );
      curr_node->next->identifier.dsc$w_length =
         strlen( curr_node->next->identifier.dsc$a_pointer );
      curr_node->next->identifier.dsc$b_dtype = DSC$K_DTYPE_T;
      curr_node->next->identifier.dsc$b_class = DSC$K_CLASS_S;

   /*------*/
   /* Get the third token, the account.                                      */

      token = strtok( NULL, K_AUTH_TOKEN_SEPARATORS );
      if( token == NULL )
         return( GIVE_AUTH_PARSERR );
      strupr( token );
      if( (curr_node->next->account.dsc$a_pointer = strdup( token ))==NULL )
         return( GIVE_AUTH_MEMERR );
      curr_node->next->account.dsc$w_length =
         strlen( curr_node->next->account.dsc$a_pointer );
      curr_node->next->account.dsc$b_dtype = DSC$K_DTYPE_T;
      curr_node->next->account.dsc$b_class = DSC$K_CLASS_S;

   /*------*/
   /* Get the fourth token, the maximum size.                                */

      token = strtok( NULL, K_AUTH_TOKEN_SEPARATORS );
      if( token == NULL )
         return( GIVE_AUTH_PARSERR );
      curr_node->next->max_size = atol( token );

   /*------*/
   /* Get the fifth and final token, the flags.                              */

      token = strtok( NULL, K_AUTH_TOKEN_SEPARATORS );
      if( token == NULL )
         return( GIVE_AUTH_PARSERR );
      strupr( token );
      if( strcmp( token, K_AUTH_TOKEN_PREFIX )==0 )
         curr_node->next->flags = K_INCOMING_PREFIX;
      else if( strcmp( token, K_AUTH_TOKEN_POSTFIX )==0 )
         curr_node->next->flags = K_INCOMING_POSTFIX;
      else if( strcmp( token, K_AUTH_TOKEN_SUBDIR )==0 )
         curr_node->next->flags = K_INCOMING_SUBDIR;
      else if( strcmp( token, K_AUTH_TOKEN_NONE )==0 )
         curr_node->next->flags = K_INCOMING_NONE;
      else
         return( GIVE_AUTH_ILLFLAG );

      curr_node = curr_node->next;
   
   }

   fclose( auth_fp );

   return( SS$_NORMAL );

}



/*---------------------------------------------------------------------------*/
/* Look_For_Authorization -- this function scans through the authorize list  */
/* attempting to match one of the entries with ourselves.                    */

ulong Look_For_Authorization( struct auth_entry *auth_head, 
                              char *directory, ulong size, char *filename,
                              struct dsc$descriptor_s username_d, 
                              struct dsc$descriptor_s account_d, 
                              struct dsc$descriptor_s *rights_d ) 
{
   ulong ss_stat;
   ulong exit_match;
   ulong exit_toobig;
   struct auth_entry *auth_curr;
   struct dsc$descriptor_s *curr_right_d;
   char root[K_FILENAME_LEN+1];
   char extension[K_FILENAME_LEN+1];

/*------*/
/* Loop through the authorization list, looking for a single record that     */
/* matches the running user's attributes.  Clear a flag that indicates, upon */
/* failure, whether we saw an authorization entry that would have been OK    */
/* except for the file size.                                                 */

   exit_toobig = FALSE;
   for( auth_curr = auth_head->next; auth_curr != NULL;
        auth_curr = auth_curr->next ) {

   /*------*/
   /* Attempt to match the username and the account name.                    */

      if( str$match_wild( &username_d, 
                          &(auth_curr->username) )==STR$_NOMATCH )
         continue;
      if( str$match_wild( &account_d, &(auth_curr->account) )==STR$_NOMATCH )
         continue;

   /*------*/
   /* Attempt to match a single identifier the user possesses.               */

      exit_match = FALSE;
      for( curr_right_d = rights_d; curr_right_d->dsc$a_pointer != NULL;
           curr_right_d++ ) {
         if( str$match_wild( curr_right_d, 
                             &(auth_curr->identifier) )==STR$_MATCH ) {
            exit_match = TRUE;
            break;
         }
      }
      if( !exit_match )
         continue;

   /*------*/
   /* Check the file size.                                                   */

      if( (auth_curr->max_size != 0) && (auth_curr->max_size < size) ) {
         exit_toobig = TRUE;
         continue;
      }

   /*------*/
   /* At this point, we know we can do the copy.  Use the flags field from   */
   /* the authorization record to modify the filename and directory provided */
   /* so that some degree of customization can be had by the destination.    */

      ss_stat = Extract_Filename( filename, root, extension );
      if( ss_failure( ss_stat ) )
         return( ss_stat );

      switch( auth_curr->flags ) {
         case K_INCOMING_PREFIX:
            strcpy( filename, username_d.dsc$a_pointer );
            strcat( filename, "-" );
            strcat( filename, root );
            strcat( filename, "." );
            strcat( filename, extension );
            break;
         case K_INCOMING_POSTFIX:
            strcpy( filename, root );
            strcat( filename, "-" );
            strcat( filename, username_d.dsc$a_pointer );
            strcat( filename, "." );
            strcat( filename, extension );
            break;
         case K_INCOMING_SUBDIR:
            directory[strlen(directory)-1] = '\0';
            strcat( directory, "." );
            strcat( directory, username_d.dsc$a_pointer );
            strcat( directory, "]" );
         default:
            strcpy( filename, root );
            strcat( filename, "." );
            strcat( filename, extension );
            break;
      }

      return( GIVE_AUTH );
   }

   if( exit_toobig )
      return( GIVE_AUTH_TOOBIG );
   else
      return( GIVE_NOAUTH );
}
